// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <fstream>
#include <sstream>
#include <vector>
#include <iterator>
#include <cstring>
#include <filesystem>
#include <pthread.h>
#include <algorithm>
#include <numeric>
#include <random>
#include <functional>
#include <sys/stat.h>
#include <sys/mman.h>
#include <math.h>
#include <thread>
#include <chrono>
#include <limits>
#include <queue>
#include <string>
#include <cctype>
#include <unordered_set>
#include "es_qwen2_client.hpp"
#include "json.h"
#include "half.h"

namespace fs = std::filesystem;

static int32_t client_log_level = 0;

#define client_log(func, line, level, fmt, ...)                                                                \
    do {                                                                                                       \
        if (level <= client_log_level)                                                                         \
            printf("[ES][Qwen2][%s](%s:%d) " fmt, level == 0 ? "ERROR" : "DEBUG", func, line, ##__VA_ARGS__); \
    } while (0)

#define ES_LOG_DBG(format, ...) client_log(__func__, __LINE__, 2, format, ##__VA_ARGS__)
#define ES_LOG_MSG(format, ...) client_log(__func__, __LINE__, 1, format, ##__VA_ARGS__)
#define ES_LOG_ERR(format, ...) client_log(__func__, __LINE__, 0, format, ##__VA_ARGS__)

template <typename T>
void check(T result, char const* const func, ES_S32 const line)
{
    if (result) {
        fprintf(stderr, "%s(%d) error=%d \n", func, line, static_cast<unsigned ES_S32>(result));
        exit(EXIT_FAILURE);
    }
}

#define esCheckError(val) check((val), __func__, __LINE__)

#define DUMP_TOKEN_PREFIX(deviceId, maskVal, layer) \
    ("device_" + std::to_string(deviceId) + "_token_" + std::to_string(maskVal) + "_layer_" + std::to_string(layer))

#define DMABUF_ALIGN_SIZE 0x200000UL  // 2M
#define ALIGNMENT_UP(size, align) (((size) + ((align)-1)) & (~((align)-1)))

std::atomic<bool> Qwen2Client::mStopped{true};
std::atomic<bool> Qwen2Client::mExited{false};

void esBindThread2Die(int32_t dieId)
{
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    if (dieId == 0) {
        CPU_SET(0, &cpuset);
        CPU_SET(1, &cpuset);
        CPU_SET(2, &cpuset);
        CPU_SET(3, &cpuset);
    } else {
        CPU_SET(4, &cpuset);
        CPU_SET(5, &cpuset);
        CPU_SET(6, &cpuset);
        CPU_SET(7, &cpuset);
    }

    pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpuset);
}

Qwen2Client::Qwen2Client()
    : mWordEmbeddings(nullptr),
    mDeviceMask(0),
    mDeviceStatus(0),
    mPostCnt(0)
{
    ES_LOG_DBG("Qwen2 Client construction done.\n");
}

void Qwen2Client::setconfig(const std::string& configPath)
{
    if (loadConfig(configPath)) {
        throw std::runtime_error("Config file is not exist.");
    }

    if (ES_SUCCESS != ES_NPU_SetLogLevel(mConfig.logLevel)) {
        ES_LOG_ERR("Set log level: %d failed\n", mConfig.logLevel);
        throw std::runtime_error("Set log level failed.");
    }

    if (loadEmbedding()) {
        throw std::runtime_error("Load embedding failed.");
    }

    if (mConfig.modelName == "qwen2") {
        mTokenizer = std::make_unique<QwenTokenizer>(mConfig.tiktokenFile, mConfig.specialTokens);
    } else if (mConfig.modelName == "deepseek_r1_distill_qwen") {
        mTokenizer = std::make_unique<DeepSeekTokenizer>(mConfig.tiktokenFile, mConfig.specialTokens);
    } else {
        ES_LOG_ERR("Not support model: %s\n", mConfig.modelName.c_str());
        throw std::runtime_error("Not support model.");
    }

    if (!mTokenizer) {
        throw std::runtime_error("Construct QwenTokenizer failed.");
    }

    mContexts.resize(mConfig.devPrivAttr.size());
    mStreams.resize(mConfig.devPrivAttr.size());
    mLoadedModels.store(0);
    mTotalModels.store(0);
    for (auto &block : mConfig.devPrivAttr) {
        mTotalModels.fetch_add(block.blockModelFiles.size());
        mTotalModels.fetch_add(mConfig.logits ? 1 : 0);
    }

    mBlockTasks.resize(mConfig.devPrivAttr.size());
    for (int32_t i = 0; i < mBlockTasks.size(); i++) {
        mBlockTasks[i].resize(mConfig.devPrivAttr[i].blockModelFiles.size());
    }
    mLogitsTask.resize(mConfig.devPrivAttr.size());

    for (int32_t i = 0; i < mConfig.devPrivAttr.size(); i++) {
        mDeviceMask.fetch_or(0x1ULL << i);
        mTokenMask.emplace_back(std::make_shared<std::atomic<int32_t>>(0));
    }
    mTopkIdx.resize(mConfig.devPrivAttr.size());
    mTopkVal.resize(mConfig.devPrivAttr.size());

    for (int32_t i = 0; i < mConfig.devPrivAttr.size(); i++) {
        mTaskThreadVec.push_back(std::thread(&Qwen2Client::taskThreadFunc, this, i));
    }

    {
        // wait load model complete
        std::unique_lock<std::mutex> syncLock(mTaskMutex);
        mCompleteCv.wait(syncLock, [this](){ return mLoadedModels.load() == mTotalModels.load() || Qwen2Client::mExited.load(); });
    }

    ES_LOG_DBG("Qwen2 Client set config done.\n");
}

Qwen2Client::~Qwen2Client() { }

void Qwen2Client::exitClient()
{
    if (Qwen2Client::mStopped.load()) {
        std::unique_lock<std::mutex> taskLock(mTaskMutex);
        Qwen2Client::mExited.store(true);
        mTaskCv.notify_all();
    } else {
        std::unique_lock<std::mutex> taskLock(mTaskMutex);
        Qwen2Client::mExited.store(true);
        Qwen2Client::mStopped.store(true);
    }

    for (auto &taskTread : mTaskThreadVec) {
        if (taskTread.joinable()) {
            taskTread.join();
        }
    }
}

ES_S32 Qwen2Client::releaseTaskBuffers(int32_t idx)
{
    MODEL_TASK_S *blockTask = &mBlockTasks[idx].front();
    if (blockTask->isPrepared) {
        esCheckError(ES_NPU_Unprepare(&blockTask->task.inputFd[IN_TOKEN_IDX]));
        esCheckError(ES_SYS_Munmap(blockTask->inputMapAddr[IN_TOKEN_IDX], blockTask->task.inputFd[IN_TOKEN_IDX].size));
        esCheckError(ES_SYS_MemFree(blockTask->task.inputFd[IN_TOKEN_IDX].memFd));

        esCheckError(ES_NPU_Unprepare(&blockTask->task.outputFd[OUT_HIDDEN_IDX]));
        esCheckError(ES_SYS_Munmap(blockTask->outputMapAddr[OUT_HIDDEN_IDX], blockTask->task.outputFd[OUT_HIDDEN_IDX].size));
        esCheckError(ES_SYS_MemFree(blockTask->task.outputFd[OUT_HIDDEN_IDX].memFd));
    } else {
        return ES_SUCCESS;
    }

    for (auto& blockTask : mBlockTasks[idx]) {
        if (blockTask.isPrepared) {
            esCheckError(ES_NPU_Unprepare(&blockTask.task.inputFd[IN_K_CACHE_IDX]));
            esCheckError(ES_SYS_Munmap(blockTask.inputMapAddr[IN_K_CACHE_IDX], blockTask.task.inputFd[IN_K_CACHE_IDX].size));
            esCheckError(ES_SYS_MemFree(blockTask.task.inputFd[IN_K_CACHE_IDX].memFd));

            esCheckError(ES_NPU_Unprepare(&blockTask.task.inputFd[IN_V_CACHE_IDX]));
            esCheckError(ES_SYS_Munmap(blockTask.inputMapAddr[IN_V_CACHE_IDX], blockTask.task.inputFd[IN_V_CACHE_IDX].size));
            esCheckError(ES_SYS_MemFree(blockTask.task.inputFd[IN_V_CACHE_IDX].memFd));

            esCheckError(ES_NPU_Unprepare(&blockTask.task.inputFd[IN_MASK_IDX]));
            esCheckError(ES_SYS_Munmap(blockTask.inputMapAddr[IN_MASK_IDX], blockTask.task.inputFd[IN_MASK_IDX].size));
            esCheckError(ES_SYS_MemFree(blockTask.task.inputFd[IN_MASK_IDX].memFd));
        } else {
            return ES_SUCCESS;
        }
    }

    if ((mConfig.logits) && (mBlockTasks[idx].back().isPrepared)) {
        esCheckError(ES_NPU_Unprepare(&mLogitsTask[idx].task.outputFd[0]));
        esCheckError(ES_SYS_Munmap(mLogitsTask[idx].outputMapAddr[0], mLogitsTask[idx].task.outputFd[0].size));
        esCheckError(ES_SYS_MemFree(mLogitsTask[idx].task.outputFd[0].memFd));
    }

    return ES_SUCCESS;
}

void Qwen2Client::printFp16Vec(const std::vector<fp16_t>& vec, int32_t num)
{
    float tmp;
    for (int32_t idx = 0; idx < num; idx++) {
        tmp = half_float::detail::half2float(vec[idx]);
        ES_LOG_DBG("%f\n", tmp);
    }
}

ES_S32 Qwen2Client::dumpData(const void* address, const size_t size, const std::string& file)
{
    std::string outputPath = std::string("dump/") + file;
    std::ofstream outputFile(outputPath, std::ios::binary);
    if (!outputFile.is_open()) {
        ES_LOG_ERR("Failed to open file: %s\n", outputPath.c_str());
        return ES_NPU_ERROR_MODEL_NOT_FOUND;
    }

    outputFile.write((const char*)address, size);
    outputFile.close();

    ES_LOG_DBG("\n===> Data dumped to file: %s\n", outputPath.c_str());
    return ES_SUCCESS;
}

ES_S32 Qwen2Client::loadConfig(const std::string& configPath)
{
    std::ifstream configFile(configPath, std::ifstream::binary);
    if (!configFile.is_open()) {
        return ES_NPU_ERROR_MODEL_NOT_FOUND;
    }

    Json::Value root;
    configFile >> root;

    const Json::Value modelName = root["model_name"];
    mConfig.modelName = modelName.asString();

    const Json::Value modelDir = root["model_dir"];
    mConfig.modelDir = modelDir.asString();

    const Json::Value logits = root["logits"];
    mConfig.logits = logits.asBool();

    mConfig.tiktokenFile = root["tokenizer_file"].asString();
    mConfig.embeddingFile = root["embedding_file"].asString();

    const Json::Value modelList = root["model_list"];
    for (const auto& elem : modelList) {
        DevicePrivAttr privAttr;
        privAttr.devId = elem["device_id"].asInt();
        privAttr.vocabOffset = elem["vocab_offset"].asInt();
        privAttr.dataDump = elem["data_dump"].asInt();
        privAttr.dataTokenInterval = CLIENT_INTERVAL_S(elem["data_dump_token"][0].asInt(), elem["data_dump_token"][1].asInt());
        privAttr.perfEnable = elem["perf_enable"].asInt();
        privAttr.opDump = elem["op_dump"].asInt();
        privAttr.opTokenInterval = CLIENT_INTERVAL_S(elem["op_dump_token"][0].asInt(), elem["op_dump_token"][1].asInt());
        privAttr.opLayerInterval = CLIENT_INTERVAL_S(elem["op_dump_layer"][0].asInt(), elem["op_dump_layer"][1].asInt());
        privAttr.opListInterval = CLIENT_INTERVAL_S(elem["op_dump_list"][0].asInt(), elem["op_dump_list"][1].asInt());
        privAttr.opType = elem["op_dump_type"].asInt();

        std::vector<std::string> modelNameList;
        for (const auto& model : elem["block_models"]) {
            privAttr.blockModelFiles.push_back(model.asString());
        }

        if (privAttr.blockModelFiles.empty()) {
            ES_LOG_ERR("There is no block model(device: %d) in %s\n", elem["device_id"].asInt(), mConfig.modelDir.c_str());
            return ES_NPU_ERROR_MODEL_NOT_FOUND;
        }

        if (mConfig.logits) {
            privAttr.logitsModelFile = elem["logits_models"].asString();
        }
        mConfig.devPrivAttr.emplace_back(privAttr);
    }

    mConfig.vocabSize = root["vocab_size"].asInt();
    mConfig.numHead = root["num_head"].asInt();
    mConfig.embeddingSize = root["embedding_size"].asInt();
    mConfig.embeddingDim = root["embedding_dim"].asInt();
    mConfig.embeddingDim1 = root["embedding_dim1"].asInt();
    mConfig.embeddingDim2 = root["embedding_dim2"].asInt();
    mConfig.tokenNum = root["token_num"].asInt();
    mConfig.tokenN1 = root["token_n1"].asInt();
    mConfig.tokenN2 = root["token_n2"].asInt();
    mConfig.precision = root["precision"].asInt();

    mConfig.language = root["language"].asString();

    const Json::Value specialTokens = root["special_tokens"];
    for (const auto& token : specialTokens) {
        mConfig.specialTokens.push_back(token.asString());
    }

    const Json::Value eosList = root["eos_list"];
    for (const auto& eos : eosList) {
        mConfig.eosList.push_back(eos.asString());
    }

    mConfig.responseMaxLen = root["response_max_len"].asInt();
    mConfig.responseLineLen = root["response_line_len"].asInt();
    mConfig.logLevel = root["log_level"].asInt();
    mConfig.randomCnt = root["random_cnt"].asInt();
    mConfig.repeatLen = root["repeat_len"].asInt();
    mConfig.repeatCnt = root["repeat_cnt"].asInt();
    mConfig.modelTest = root["model_test"].asInt();
    mConfig.repetitionPenalty = root["repetition_penalty"].asFloat();
    mConfig.topK = root["top_k"].asInt();

    const Json::Value& prompts = root["prompts"];
    for (Json::ValueConstIterator it = prompts.begin(); it != prompts.end(); ++it) {
        int32_t key = std::stoi(it.key().asString());
        mConfig.prompts[key] = it->asString();
    }

    const Json::Value& rolePrompts = root["role_prompts"];
    for (Json::ValueConstIterator it = rolePrompts.begin(); it != rolePrompts.end(); ++it) {
        int32_t key = std::stoi(it.key().asString());
        mConfig.rolePrompts[key] = it->asString();
    }

    client_log_level = mConfig.logLevel;

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::loadEmbedding()
{
    try {
        mWordEmbeddings = new fp16_t[mConfig.vocabSize * mConfig.embeddingSize];
    } catch (const std::bad_alloc& e) {
        ES_LOG_ERR("alloc memory for word embedding failed\n");
        return ES_FAILURE;
    }

    std::ifstream file(mConfig.embeddingFile, std::ios::binary | std::ios::ate);
    if (!file.is_open()) {
        ES_LOG_ERR("Failed to open embedding file: %s\n",mConfig.embeddingFile.c_str());
        return ES_FAILURE;
    }

    std::streampos fileSize = file.tellg();
    if (static_cast<int>(fileSize) != mConfig.vocabSize * mConfig.embeddingSize * sizeof(fp16_t)) {
        ES_LOG_ERR("Invalid embedding file size: %d\n", static_cast<int>(fileSize));
        return ES_FAILURE;
    }

    file.seekg(0, std::ios::beg);
    file.read(reinterpret_cast<char*>(mWordEmbeddings), fileSize);

    if (!file) {
        ES_LOG_ERR("Error reading embedding file.\n");
        return ES_FAILURE;
    }

    file.close();

    return ES_SUCCESS;
}

void Qwen2Client::displayLoadModelProgress(int32_t current, int32_t total, double costTime)
{
    float progress = (float)current / total * 100.0f;
    std::string progressBar;

    progressBar = "\rLoading models: [";
    int pos = 50 * progress / 100.0;

    for (int i = 0; i < 50; ++i) {
        if (i < pos)
            progressBar += "=";
        else if (i == pos)
            progressBar += ">";
        else
            progressBar += " ";
    }

    progressBar += "] ";

    std::ostringstream progressFmt;
    progressFmt << std::setw(6) << std::setfill(' ') << std::fixed << std::setprecision(2) << progress;
    progressBar += progressFmt.str() + "%" + " ( " + std::to_string(costTime) + " seconds )";

    std::cout << progressBar;

    std::cout.flush();
}

ES_S32 Qwen2Client::initDevice(int32_t idx)
{
    uint16_t deviceNum = 0;

    esCheckError(ES_NPU_GetNumDevices(&deviceNum));
    if (mConfig.devPrivAttr[idx].devId >= deviceNum) {
        ES_LOG_ERR("deviceId(%d) is not less than deviceNum(%d)\n", mConfig.devPrivAttr[idx].devId, deviceNum);
        return ES_NPU_ERROR_DEV_NOT_FOUND;
    }

    ES_LOG_DBG("open device:%u\n", mConfig.devPrivAttr[idx].devId);
    esCheckError(ES_NPU_SetDevice(mConfig.devPrivAttr[idx].devId));

    esCheckError(ES_NPU_CreateContext(&mContexts[idx], mConfig.devPrivAttr[idx].devId));

    esCheckError(ES_NPU_CreateStream(&mStreams[idx]));

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::deinitDevice(int32_t idx)
{
    esCheckError(ES_NPU_DestroyStream(mStreams[idx]));
    esCheckError(ES_NPU_DestroyContext(mContexts[idx]));
    esCheckError(ES_NPU_ReleaseDevice(mConfig.devPrivAttr[idx].devId));
    return ES_SUCCESS;
}

ES_S32 Qwen2Client::loadModel(int32_t idx)
{
    uint32_t modelId;

    auto start = std::chrono::high_resolution_clock::now();
    // load  block model
    for (int32_t num = 0; num < mConfig.devPrivAttr[idx].blockModelFiles.size(); num++) {
        if (Qwen2Client::mExited.load()) {
            return ES_SUCCESS;
        }
        std::string model = mConfig.modelDir + "/" + mConfig.devPrivAttr[idx].blockModelFiles[num];

        esCheckError(ES_NPU_LoadModelFromFile(&modelId, model.c_str()));
        ES_LOG_DBG("load %d model:%s, modelId=%u\n", num, model.c_str(), modelId);

        esCheckError(prepareBlockTaskBuffers(idx, num, modelId));

        auto end = std::chrono::high_resolution_clock::now();
        std::chrono::duration<double> elapsed = end - start;
        displayLoadModelProgress(mLoadedModels.load() + 1, mTotalModels.load(), elapsed.count());
        mLoadedModels.fetch_add(1);
    }

    // load logits model
    if (mConfig.logits) {
        std::string model = mConfig.modelDir + "/" + mConfig.devPrivAttr[idx].logitsModelFile;
        esCheckError(ES_NPU_LoadModelFromFile(&modelId, model.c_str()));
        ES_LOG_DBG("load logits model:%s, modelId=%u\n", model.c_str(), modelId);

        esCheckError(prepareLogitsTaskBuffers(idx, modelId));
        auto end = std::chrono::high_resolution_clock::now();
        std::chrono::duration<double> elapsed = end - start;
        displayLoadModelProgress(mLoadedModels.load() + 1, mTotalModels.load(), elapsed.count());
        mLoadedModels.fetch_add(1);
    }

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::unloadModel(int32_t idx)
{
    for (auto& blockTask : mBlockTasks[idx]) {
        if (blockTask.isPrepared) {
            esCheckError(ES_NPU_UnloadModel(blockTask.task.modelId));
        } else {
            break;
        }
    }

    if ((mConfig.logits) && (mBlockTasks[idx].back().isPrepared)) {
        esCheckError(ES_NPU_UnloadModel(mLogitsTask[idx].task.modelId));
    }

    releaseTaskBuffers(idx);

    if (idx == 0) {
        delete[] mWordEmbeddings;
        mWordEmbeddings = nullptr;
    }

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareTensorBuffer(NPU_TENSOR_S &tensor, ES_DEV_BUF_S &devBuf, void **pAddr, uint16_t devId, bool cache)
{
    ES_U64 memFd;
    SYS_CACHE_MODE_E cacheMode = cache ? SYS_CACHE_MODE_CACHED : SYS_CACHE_MODE_NOCACHE;
    const char *zone = (devId == 0) ? "mmz_nid_0_part_0" : "mmz_nid_1_part_0";
    ES_U64 bufferSize;

    bufferSize = (cache) ? tensor.bufferSize : ALIGNMENT_UP(tensor.bufferSize, DMABUF_ALIGN_SIZE);
    esCheckError(ES_SYS_MemAlloc(&memFd, cacheMode, "qwen2_demo", zone, bufferSize));
    *pAddr = ES_SYS_Mmap(memFd, tensor.bufferSize, cacheMode);
    if (!*pAddr) {
        ES_LOG_ERR("ES_SYS_Mmap failed!\n");
        return ES_FAILURE;
    }

    (void)memset(*pAddr, 0x00, tensor.bufferSize);
    if (cache) {
        esCheckError(ES_SYS_MemFlushCache(memFd));
    }

    devBuf.size = tensor.bufferSize;
    devBuf.offset = 0;
    devBuf.memFd = memFd;

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareKVCacheBuffers(uint32_t modelId, MODEL_TASK_S &modelTask, uint16_t devId)
{
    void* pData;
    ES_U64 memFd;
    ES_U64 kCacheSize = mConfig.precision * mConfig.numHead * mConfig.embeddingDim1 * mConfig.tokenNum * mConfig.embeddingDim2;
    ES_U64 vCacheSize = mConfig.precision * mConfig.numHead * mConfig.tokenN1 * mConfig.embeddingDim * mConfig.tokenN2;
    ES_U64 maskSize = MASK_BUFFER_SIZE;
    NPU_TENSOR_S tensor;

    // Prepare K cache buffer
    esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_K_CACHE_IDX, &tensor));
    if (tensor.bufferSize != kCacheSize) {
        ES_LOG_ERR("k cache size is not valid, %d!\n", tensor.bufferSize);
        return ES_FAILURE;
    }
    esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_K_CACHE_IDX],
                                     &modelTask.inputMapAddr[IN_K_CACHE_IDX], devId, false));

    modelTask.task.outputFd[OUT_K_CACHE_IDX] = modelTask.task.inputFd[IN_K_CACHE_IDX];
    modelTask.outputMapAddr[OUT_K_CACHE_IDX] = modelTask.inputMapAddr[IN_K_CACHE_IDX];
    esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_K_CACHE_IDX]));

    // Prepare V cache buffer
    esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_V_CACHE_IDX, &tensor));
    if (tensor.bufferSize != vCacheSize) {
        ES_LOG_ERR("v cache size is not valid, %d!\n", tensor.bufferSize);
        return ES_FAILURE;
    }
    esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_V_CACHE_IDX],
                                     &modelTask.inputMapAddr[IN_V_CACHE_IDX], devId, false));

    modelTask.task.outputFd[OUT_V_CACHE_IDX] = modelTask.task.inputFd[IN_V_CACHE_IDX];
    modelTask.outputMapAddr[OUT_V_CACHE_IDX] = modelTask.inputMapAddr[IN_V_CACHE_IDX];
    esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_V_CACHE_IDX]));

    ES_LOG_DBG("prepare KVCache buffers done.\n");

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareBlockTaskBuffers(int32_t idx, int32_t modelIdx, uint32_t modelId)
{
    MODEL_TASK_S &modelTask = mBlockTasks[idx][modelIdx];
    NPU_TENSOR_S tensor;
    int32_t numInputTensors, numOutputTensors;

    memset(&modelTask, 0, sizeof(MODEL_TASK_S));
    esCheckError(ES_NPU_GetNumInputTensors(modelId, &numInputTensors));
    if (numInputTensors != IN_MAX_IDX) {
        ES_LOG_ERR("Input tensor num is invalid from model, numInputTensors=%d\n", numInputTensors);
        return ES_NPU_ERROR_INVALID_SIZE;
    }
    modelTask.task.inputFdNum = numInputTensors;

    esCheckError(ES_NPU_GetNumOutputTensors(modelId, &numOutputTensors));
    if (numOutputTensors != OUT_MAX_IDX) {
        ES_LOG_ERR("Output tensor num is invalid from model, numOutputTensors=%d\n", numOutputTensors);
        return ES_NPU_ERROR_INVALID_SIZE;
    }
    modelTask.task.outputFdNum = numOutputTensors;

    // prepare KV Cache
    esCheckError(prepareKVCacheBuffers(modelId, modelTask, mConfig.devPrivAttr[idx].devId));

    // preapre mask buffer
    esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_MASK_IDX, &tensor));
    esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_MASK_IDX],
                                     &modelTask.inputMapAddr[IN_MASK_IDX], mConfig.devPrivAttr[idx].devId, false));
    esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_MASK_IDX]));

    // prepare token buffer
    if (modelIdx == 0) {
        esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_TOKEN_IDX, &tensor));
        esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_TOKEN_IDX],
                                         &modelTask.inputMapAddr[IN_TOKEN_IDX], mConfig.devPrivAttr[idx].devId, false));
        esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_TOKEN_IDX]));

        esCheckError(ES_NPU_GetOutputTensorDesc(modelId, OUT_HIDDEN_IDX, &tensor));
        esCheckError(prepareTensorBuffer(tensor, modelTask.task.outputFd[OUT_HIDDEN_IDX],
                                         &modelTask.outputMapAddr[OUT_HIDDEN_IDX], mConfig.devPrivAttr[idx].devId, false));
        esCheckError(ES_NPU_Prepare(&modelTask.task.outputFd[OUT_HIDDEN_IDX]));
    } else {
        modelTask.task.inputFd[IN_TOKEN_IDX] = mBlockTasks[idx][modelIdx - 1].task.outputFd[OUT_HIDDEN_IDX];
        modelTask.inputMapAddr[IN_TOKEN_IDX] = mBlockTasks[idx][modelIdx - 1].outputMapAddr[OUT_HIDDEN_IDX];

        modelTask.task.outputFd[OUT_HIDDEN_IDX] = mBlockTasks[idx][modelIdx - 1].task.inputFd[IN_TOKEN_IDX];
        modelTask.outputMapAddr[OUT_HIDDEN_IDX] = mBlockTasks[idx][modelIdx - 1].inputMapAddr[IN_TOKEN_IDX];
    }

    modelTask.task.modelId = modelId;
    modelTask.isPrepared = true;

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareLogitsTaskBuffers(int32_t idx, uint32_t modelId)
{
    NPU_TENSOR_S tensor;
    int32_t numInputTensors, numOutputTensors;

    memset(&mLogitsTask[idx], 0x00, sizeof(MODEL_TASK_S));
    esCheckError(ES_NPU_GetNumInputTensors(modelId, &numInputTensors));
    if (numInputTensors > 1) {
        ES_LOG_ERR("logits task numInput tensor can not align: %d\n", numInputTensors);
    }

    mLogitsTask[idx].task.inputFdNum = numInputTensors;
    mLogitsTask[idx].task.inputFd[0] = mBlockTasks[idx].back().task.outputFd[OUT_HIDDEN_IDX];
    mLogitsTask[idx].inputMapAddr[0] = mBlockTasks[idx].back().inputMapAddr[OUT_HIDDEN_IDX];

    esCheckError(ES_NPU_GetNumOutputTensors(modelId, &numOutputTensors));
    mLogitsTask[idx].task.outputFdNum = numOutputTensors;
    for (int32_t i = 0; i < numOutputTensors; i++) {
        esCheckError(ES_NPU_GetOutputTensorDesc(modelId, i, &tensor));
        esCheckError(prepareTensorBuffer(tensor, mLogitsTask[idx].task.outputFd[i], &mLogitsTask[idx].outputMapAddr[i], mConfig.devPrivAttr[idx].devId, true));
        esCheckError(ES_NPU_Prepare(&mLogitsTask[idx].task.outputFd[i]));
    }

    mLogitsTask[idx].task.modelId = modelId;

    return ES_SUCCESS;
}

ES_U64 Qwen2Client::calcKCacheOutputAddrOffset(int32_t num)
{
    ES_U64 offset;

    return mConfig.precision * mConfig.embeddingDim2 * num;
}

ES_U64 Qwen2Client::calcVCacheOutputAddrOffset(int32_t num)
{
    int32_t dim0 = num / mConfig.tokenN2;
    int32_t dim1 = num % mConfig.tokenN2;

    return mConfig.precision * (mConfig.tokenN2 * mConfig.embeddingDim * dim0 + dim1);
}

void Qwen2Client::runModel(MODEL_TASK_S& modelTask)
{
    // Submit model task
    esCheckError(ES_NPU_Submit(&modelTask.task, 1));
}

void Qwen2Client::runModelAsync(MODEL_TASK_S& modelTask, npu_stream stream)
{
    // Submit model task async
    esCheckError(ES_NPU_SubmitAsync(&modelTask.task, 1, stream));
}

int32_t Qwen2Client::argmax(int16_t *data, int32_t size, int16_t *max)
{
    int32_t pos = 0;
    *max = data[0];

    for (int32_t i = 1; i < size; i++) {
        if (data[i] > *max) {
            pos = i;
            *max = data[i];
        }
    }

    return pos;
}

prob_data_t Qwen2Client::getTopK(int32_t idx, int32_t wordPosInVocab, const int32_t& maskVal, const std::vector<int32_t>& outputIds)
{
    prob_data_t topK;
    std::chrono::_V2::system_clock::time_point startTime, modelStartTime, lmStartTime, postStartTime, endTime;

    if (mConfig.devPrivAttr[idx].perfEnable) {
        startTime = std::chrono::high_resolution_clock::now();
    }

    // Get embeddings for next word
    int32_t wordStart = wordPosInVocab * mConfig.embeddingSize;
    int32_t wordEnd = wordStart + mConfig.embeddingSize;
    std::vector<fp16_t> wordEmbedding(mWordEmbeddings + wordStart, mWordEmbeddings + wordEnd);

    // prepare input token
    memcpy(mBlockTasks[idx].front().inputMapAddr[IN_TOKEN_IDX],
           wordEmbedding.data(), wordEmbedding.size() * sizeof(fp16_t));

    if (mConfig.devPrivAttr[idx].dataDump && checkInterval(maskVal, mConfig.devPrivAttr[idx].dataTokenInterval)) {
        std::string fileName = "token_" + std::to_string(maskVal) + std::string("_wordembedding.data");
        dumpData(mBlockTasks[idx].front().inputMapAddr[IN_TOKEN_IDX], mBlockTasks[idx].front().task.inputFd[IN_TOKEN_IDX].size, fileName);
    }

    if (mConfig.devPrivAttr[idx].perfEnable) {
        modelStartTime = std::chrono::high_resolution_clock::now();
    }

    // run block models
    for (int32_t modelIdx = 0; modelIdx < mBlockTasks[idx].size(); modelIdx++) {
        // adjust output KVCache offset
        mBlockTasks[idx][modelIdx].task.outputFd[OUT_K_CACHE_IDX].offset = calcKCacheOutputAddrOffset(maskVal);
        mBlockTasks[idx][modelIdx].task.outputFd[OUT_V_CACHE_IDX].offset = calcVCacheOutputAddrOffset(maskVal);
        // fill mask
        *static_cast<int16_t*>(mBlockTasks[idx][modelIdx].inputMapAddr[IN_MASK_IDX]) = maskVal;

        ES_LOG_DBG("start runmodel mask=%d, modelIdx=%d, modelId: %d\n", maskVal, modelIdx, mBlockTasks[idx][modelIdx].task.modelId);
        if (mConfig.devPrivAttr[idx].opDump &&
            checkInterval(maskVal, mConfig.devPrivAttr[idx].opTokenInterval) &&
            checkInterval(modelIdx, mConfig.devPrivAttr[idx].opLayerInterval)) {
            NPU_DUMP_CONF_S dumpConf;
            std::string dumpPath = "/tmp/" + std::to_string(idx) + "_" + std::to_string(maskVal) + "_" + std::to_string(modelIdx);
            char dumpStr[64] = {0};

            snprintf(dumpStr, 64, "%d-%d", mConfig.devPrivAttr[idx].opListInterval.min, mConfig.devPrivAttr[idx].opListInterval.max);
            dumpConf.subModelId = mConfig.devPrivAttr[idx].opType;
            dumpConf.dumpList = dumpStr;

            esCheckError(ES_NPU_ConfigDumpInfo(mBlockTasks[idx][modelIdx].task.modelId, dumpPath.c_str(), &dumpConf, 1));
            esCheckError(ES_NPU_EnableDump(mBlockTasks[idx][modelIdx].task.modelId));
        }

        // run block model
        if (mConfig.devPrivAttr[idx].opDump || mConfig.devPrivAttr[idx].dataDump) {
            runModel(mBlockTasks[idx][modelIdx]);
            if (mConfig.devPrivAttr[idx].opDump &&
                checkInterval(maskVal, mConfig.devPrivAttr[idx].opTokenInterval) &&
                checkInterval(modelIdx, mConfig.devPrivAttr[idx].opLayerInterval)) {
                esCheckError(ES_NPU_DisableDump(mBlockTasks[idx][modelIdx].task.modelId));
            }
            if (mConfig.devPrivAttr[idx].dataDump && checkInterval(maskVal, mConfig.devPrivAttr[idx].dataTokenInterval)) {
                std::string fileName = DUMP_TOKEN_PREFIX(idx, maskVal, modelIdx) + std::string("_token.data");
                dumpData(mBlockTasks[idx][modelIdx].outputMapAddr[OUT_HIDDEN_IDX],
                        mBlockTasks[idx][modelIdx].task.outputFd[OUT_HIDDEN_IDX].size, fileName);
                fileName = DUMP_TOKEN_PREFIX(idx, maskVal, modelIdx) + std::string("_kcache.data");
                dumpData(mBlockTasks[idx][modelIdx].outputMapAddr[OUT_K_CACHE_IDX],
                        mBlockTasks[idx][modelIdx].task.outputFd[OUT_K_CACHE_IDX].size, fileName);
                fileName = DUMP_TOKEN_PREFIX(idx, maskVal, modelIdx) + std::string("_vcache.data");
                dumpData(mBlockTasks[idx][modelIdx].outputMapAddr[OUT_V_CACHE_IDX],
                        mBlockTasks[idx][modelIdx].task.outputFd[OUT_V_CACHE_IDX].size, fileName);
            }
        } else {
            runModelAsync(mBlockTasks[idx][modelIdx], mStreams[idx]);
        }
    }

    if (mConfig.devPrivAttr[idx].perfEnable) {
        esCheckError(ES_NPU_SynchronizeStream(mStreams[idx]));
        lmStartTime = std::chrono::high_resolution_clock::now();
    }

    // run logits model
    if (mConfig.logits) {
        if (mConfig.devPrivAttr[idx].dataDump && checkInterval(maskVal, mConfig.devPrivAttr[idx].dataTokenInterval)) {
            runModel(mLogitsTask[idx]);
            std::string fileName = "device" + std::to_string(idx) + "_token_" + std::to_string(maskVal) + std::string("_logits_out.data");
            dumpData(mLogitsTask[idx].outputMapAddr[0], mLogitsTask[idx].task.outputFd[0].size, fileName);
        } else {
            runModelAsync(mLogitsTask[idx], mStreams[idx]);
        }
    }
    esCheckError(ES_NPU_SynchronizeStream(mStreams[idx]));

    if (mConfig.devPrivAttr[idx].perfEnable) {
        postStartTime = std::chrono::high_resolution_clock::now();
    }

    // postprocess
    if (mConfig.repetitionPenalty != 1.0) {
        // logits preprocess
        std::unordered_set<int> uniqueOutputIds(outputIds.begin(), outputIds.end());
        uint16_t *data = (uint16_t*)mLogitsTask[idx].outputMapAddr[0];

        for (auto ids : uniqueOutputIds) {
            // vocab ids > vocab offset
            if (ids < mConfig.devPrivAttr[idx].vocabOffset) {
                continue;
            }

            // vocab ids < vocab offset + vocab nums
            if (ids > (mConfig.devPrivAttr[idx].vocabOffset + mLogitsTask[idx].task.outputFd[0].size / sizeof(fp16_t))) {
                continue;
            }

            float tmp = half_float::detail::half2float(data[ids - mConfig.devPrivAttr[idx].vocabOffset]);
            if (tmp > 0) {
                tmp /= mConfig.repetitionPenalty;
            } else {
                tmp *= mConfig.repetitionPenalty;
            }
            data[ids - mConfig.devPrivAttr[idx].vocabOffset] = half_float::detail::float2half<std::round_to_nearest>(tmp);
        }
    }

    int16_t maxVal;
    topK = argmax((int16_t*)mLogitsTask[idx].outputMapAddr[0], mLogitsTask[idx].task.outputFd[0].size / sizeof(fp16_t), &maxVal);
    esCheckError(ES_SYS_MemFlushCache(mLogitsTask[idx].task.outputFd[0].memFd));

    {
        // notify the status to other thread.
        std::unique_lock<std::mutex> postSynclock(mPostSyncMutex);
        mDeviceStatus.fetch_or(0x1ULL << idx);
        mTopkVal[idx] = maxVal;
        mTopkIdx[idx] = topK;
        mPostSyncCv.notify_all();
    }

    {
        std::unique_lock<std::mutex> postSynclock(mPostSyncMutex);
        mPostSyncCv.wait(postSynclock, [this]() { return !(mDeviceStatus.load() ^ mDeviceMask.load()) || Qwen2Client::mStopped.load(); });
        mPostCnt.fetch_add(1);
        // clear device status when all device finished
        if (mPostCnt.load() == mConfig.devPrivAttr.size()) {
            mPostCnt.store(0);
            mDeviceStatus.store(0);
        }
    }
    // merge the result of all thread.
    auto maxIt = std::max_element(mTopkVal.begin(), mTopkVal.end());
    int32_t maxIndex = std::distance(mTopkVal.begin(), maxIt);
    topK = mTopkIdx[maxIndex] + mConfig.devPrivAttr[maxIndex].vocabOffset;

    if (mConfig.devPrivAttr[idx].perfEnable) {
        endTime = std::chrono::high_resolution_clock::now();

        double preprocessTime = std::chrono::duration_cast<std::chrono::microseconds>(modelStartTime - startTime).count() / 1000.0;
        double modelTime = std::chrono::duration_cast<std::chrono::microseconds>(lmStartTime - modelStartTime).count() / 1000.0;
        double lmTime = std::chrono::duration_cast<std::chrono::microseconds>(postStartTime - lmStartTime).count() / 1000.0;
        double postprocessTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime - postStartTime).count() / 1000.0;
        double tokenTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime - startTime).count() / 1000.0;
        double postTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime - postStartTime).count() / 1000.0;
        printf("dev_idx: %d, token: %d, preprocess: %fms, model: %fms, lm: %fms, postTime: %fms, time: %fms\n",
               idx, maskVal, preprocessTime, modelTime, lmTime, postTime, tokenTime);
    }

    ES_LOG_MSG("mask=%d topK=%hu\n", maskVal, topK);

    // Return topK value
    return topK;
}

std::vector<std::string> Qwen2Client::splitUtf8(const std::string& str)
{
    std::vector<std::string> result;
    std::string charBuffer;

    const size_t UTF8_1BYTE = 0x80;
    const size_t UTF8_2BYTES_LEAD = 0x6;
    const size_t UTF8_3BYTES_LEAD = 0xe;
    const size_t UTF8_4BYTES_LEAD = 0x1e;

    for (std::size_t i = 0; i < str.size();) {
        charBuffer.clear();
        unsigned char lead = str[i];
        size_t charLen = 0;

        if (lead < UTF8_1BYTE) {
            charLen = 1;
        } else if ((lead >> 5) == UTF8_2BYTES_LEAD) {
            charLen = 2;
        } else if ((lead >> 4) == UTF8_3BYTES_LEAD) {
            charLen = 3;
        } else if ((lead >> 3) == UTF8_4BYTES_LEAD) {
            charLen = 4;
        } else {
            ++i;
            continue;
        }

        charBuffer = str.substr(i, charLen);
        result.push_back(charBuffer);
        ES_LOG_MSG("split word:%s\n", charBuffer.c_str());
        i += charLen;
    }

    return result;
}

std::vector<std::string> Qwen2Client::splitEn(const std::string& input)
{
    std::vector<std::string> tokens;
    std::stringstream ss(input);
    std::string token;
    while (ss >> token) {
        std::string word;
        std::string punctuation;
        for (char c : token) {
            if (ispunct(static_cast<unsigned char>(c))) {
                if (!word.empty()) {
                    tokens.push_back(word);
                    word.clear();
                }
                punctuation += c;
                tokens.push_back(punctuation);
                punctuation.clear();
            } else {
                word += c;
            }
        }
        if (!word.empty()) {
            tokens.push_back(word);
        }
    }

    for (auto& wordStr : tokens) {
        ES_LOG_MSG("wordStr=%s\n", wordStr.c_str());
    }

    return tokens;
}

std::vector<std::string> Qwen2Client::splitSentence(const std::string& input)
{
    if (mConfig.language == "zh") {
        return splitUtf8(input);
    } else {
        return splitEn(input);
    }
}

fp16_t Qwen2Client::fp16Add(fp16_t a, fp16_t b)
{
    float aFloat = half_float::detail::half2float(a);
    float bFloat = half_float::detail::half2float(b);
    float sumFloat = aFloat + bFloat;

    fp16_t result = half_float::detail::float2half<std::round_to_nearest>(sumFloat);

    return result;
}

bool Qwen2Client::checkDuplicateWords(const std::string& sentence)
{
    std::vector<std::string> words = splitSentence(sentence);
    if (words.size() < mConfig.repeatLen * 2) {
        return false;
    }

    std::unordered_map<std::string, int32_t> sequenceCounts;

    for (size_t i = 0; i <= words.size() - mConfig.repeatLen; ++i) {
        std::string subString;
        for (size_t j = i; j < i + mConfig.repeatLen; ++j) {
            subString += words[j];
        }
        sequenceCounts[subString]++;

        if (sequenceCounts[subString] == mConfig.repeatCnt) {
            ES_LOG_MSG("\nduplicate str:%s\n", subString.c_str());
            return true;
        }
    }

    return false;
}

ES_S32 Qwen2Client::taskThreadFunc(int32_t idx)
{
    ES_S32 ret;
    bool master = (idx == 0) ? true : false;

    esBindThread2Die(mConfig.devPrivAttr[idx].devId);

    ret = initDevice(idx);
    if (ES_SUCCESS != ret) {
        ES_LOG_ERR("init device(%d) error: 0x%x\n", mConfig.devPrivAttr[idx].devId, ret);
        return ES_FAILURE;
    }

    ret = loadModel(idx);
    if (ES_SUCCESS != ret) {
        throw std::runtime_error("Load model error.");
    }

    if (mLoadedModels.load() == mTotalModels.load() || Qwen2Client::mExited.load()) {
        std::unique_lock<std::mutex> syncLock(mTaskMutex);
        mCompleteCv.notify_one();
    }

    while (!Qwen2Client::mExited.load()) {
        {
            std::unique_lock<std::mutex> syncLock(mTaskMutex);
            mTaskCv.wait(syncLock, []() { return Qwen2Client::mExited.load() || !Qwen2Client::mStopped.load(); });
            if (Qwen2Client::mExited.load()) {
                break;
            }
        }

        mTokenMask[idx]->store(0);
        for (int32_t i = 0; i < mBlockTasks[idx].size(); i++) {
            memset(mBlockTasks[idx][i].inputMapAddr[IN_V_CACHE_IDX], 0x00, mBlockTasks[idx][i].task.inputFd[IN_V_CACHE_IDX].size);
            memset(mBlockTasks[idx][i].inputMapAddr[IN_K_CACHE_IDX], 0x00, mBlockTasks[idx][i].task.inputFd[IN_K_CACHE_IDX].size);
        }

        std::string response = generate(idx, master);

        if (!Qwen2Client::mStopped.load()) {
            Qwen2Client::mStopped.store(true);
        }
        if (master) {
            std::unique_lock<std::mutex> syncLock(mTaskMutex);
            mCompleteCv.notify_all();
        }
    }

    unloadModel(idx);
    deinitDevice(idx);

    return ES_SUCCESS;
}

inline bool Qwen2Client::checkInterval(int32_t value, const CLIENT_INTERVAL_S &interval)
{
    return (value >= interval.min && value <= interval.max);
}

std::string Qwen2Client::generate(int32_t idx, bool master)
{
    std::string newSentence = "";

    std::vector<int32_t> outputIds(mSentencePosInVocab);
    int32_t wordPosInVocab = mSentencePosInVocab[0];
    std::string inputWord;
    bool endFlag = false;
    int32_t lineLength = 0;
    int32_t i;

    auto startTime = std::chrono::high_resolution_clock::now();

    ES_LOG_MSG("mSentencePosInVocab size=%u\n", mSentencePosInVocab.size());
    if (mSentencePosInVocab.size() == 0) {
        ES_LOG_ERR("master: %d, input sentence is empty, no response\n", master);
        return newSentence;
    }

    for (i = 0; i < mConfig.responseMaxLen; ++i) {
        inputWord = mTokenizer->decode(std::vector<int32_t>(1, wordPosInVocab));

        // stop to generate
        if (i >= mSentencePosInVocab.size()) {
            for (const auto& endStr : mConfig.eosList) {
                if (endStr == inputWord) {
                    ES_LOG_MSG("遇到结束词%s, 生成过程结束。", endStr.c_str());
                    endFlag = true;
                    break;
                }
            }
            if (endFlag) {
                // add the word count of eos
                mTokenMask[idx]->fetch_add(1);
                break;
            }
        }

        // Skip the original input sentence and print response one by one.
        if (i >= mSentencePosInVocab.size()) {
            outputIds.emplace_back(wordPosInVocab);
            if (master) {
                std::cout << inputWord << std::flush;
            }
        }

        if (lineLength >= mConfig.responseLineLen) {
            if (mConfig.language == "zh") {
                std::cout << std::endl;
            }
            lineLength = 0;
        }
        lineLength += inputWord.size();
        newSentence += inputWord;

        // if (checkDuplicateWords(newSentence)) {
        //     goto gen_done;
        // }
        ES_LOG_MSG("Input sentence_pos(%d) vocab_pos(%d): %s\n", i, wordPosInVocab, inputWord.c_str());

        // Get next word token with top_k value, skip the original input sentence.
        wordPosInVocab = getTopK(idx, wordPosInVocab, i, outputIds);

        if (i < mSentencePosInVocab.size() - 1) {
            wordPosInVocab = mSentencePosInVocab[i + 1];
        }
        ES_LOG_MSG("Output mask(%d) sentence_pos(%d) vocab_pos(%d)\n", i, i + 1, wordPosInVocab);

        mTokenMask[idx]->fetch_add(1);

        // check user stop generate or quit it.
        if (Qwen2Client::mStopped.load()) {
            auto maxIt = std::max_element(mTokenMask.begin(), mTokenMask.end(), [](const auto& a, const auto& b) {
                return a->load() < b->load();
            });
            if (mTokenMask[idx]->load() >= (*maxIt)->load()) {
                break;
            }
        }
    }

    auto endTime = std::chrono::high_resolution_clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime - startTime).count() / 1000.0;
    double tps = 1000 / (totalTime / mTokenMask[idx]->load());

    if (master) {
        std::cout << std::endl <<
                     "--------------------------------------------------------" << std::endl <<
                     "Throughput: " << tps << "tokens/s" << std::endl <<
                     "--------------------------------------------------------" << std::endl;
    }

    return newSentence;
}

std::map<int32_t, std::string> Qwen2Client::getOptionalPrompts()
{
    std::map<int32_t, std::string> randomMap;

    if (mConfig.randomCnt >= mConfig.prompts.size()) {
        return mConfig.prompts;
    }

    std::vector<std::map<int32_t, std::string>::const_iterator> iters;
    for (auto it = mConfig.prompts.begin(); it != mConfig.prompts.end(); ++it) {
        iters.push_back(it);
    }

    std::random_device rd;
    std::mt19937 g(rd());
    std::shuffle(iters.begin(), iters.end(), g);

    for (int32_t i = 0; i < mConfig.randomCnt; ++i) {
        randomMap[i] = iters[i]->second;
    }

    return randomMap;
}

std::map<int32_t, std::string> Qwen2Client::getOptionalRolePrompts()
{
    return mConfig.rolePrompts;
}

std::string Qwen2Client::genResponse(const std::string& rolePrompt, const std::string& prompts)
{
    std::vector<ChatMessage> messages;

    messages.emplace_back(ChatMessage(ChatMessage::ROLE_SYSTEM, rolePrompt));
    messages.emplace_back(ChatMessage(ChatMessage::ROLE_USER, prompts));

    auto startTime = std::chrono::high_resolution_clock::now();

    mSentencePosInVocab = mTokenizer->encode_messages(messages, mConfig.responseMaxLen);
    {
        std::unique_lock<std::mutex> syncLock(mTaskMutex);
        Qwen2Client::mStopped.store(false);
        mTaskCv.notify_all();
    }

    {
        std::unique_lock<std::mutex> syncLock(mTaskMutex);
        mCompleteCv.wait(syncLock, []() { return Qwen2Client::mStopped.load() || Qwen2Client::mExited.load(); });
    }

    auto endTime = std::chrono::high_resolution_clock::now();
    double duration = std::chrono::duration_cast<std::chrono::microseconds>(endTime - startTime).count();

    if (mConfig.devPrivAttr[0].perfEnable) {
        std::cout << "Times:" << (duration / 1000.0 / 1000.0) << "(s)" << std::endl;
    }

    return "";
}
