// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <csignal>
#include <iostream>
#include <sstream>
#include <string>
#include <limits>
#include <thread>
#include <chrono>
#include <locale>
#include <unistd.h>
#include <sys/resource.h>
#include "es_qwen2_client.hpp"

#if BUILD_RUN_LLM == 1
#define USER_INPUT ": 自定义输入"
#else
#define USER_INPUT ": Customized prompts"
#endif

volatile std::sig_atomic_t signalReceived = 0;

std::shared_ptr<Qwen2Client> gQwen2Client;

void signalHandler(int signal)
{
    // signalReceived = signal;
    if (signal == SIGINT) {
        gQwen2Client->exitClient();
        std::freopen("/dev/null", "r", stdin);
    } else if (signal == SIGTSTP) {
        std::cout << "\nQwen2 停止生成" << std::endl;
        Qwen2Client::mStopped.store(true);
    }
}

std::string getChineseInput()
{
    std::locale::global(std::locale(""));
    std::wstring winput;
    std::getline(std::wcin, winput);
    std::string input(winput.begin(), winput.end());

    for (char c : input) {
        if (!std::isalpha(c) || !std::isalpha(c, std::locale("zh_CN.UTF-8"))) {
            return "";
        }
    }

    return input;
}

int setFdLimit(int limit)
{
    struct rlimit rl;
    if (getrlimit(RLIMIT_NOFILE, &rl) == -1) {
        perror("getrlimit failed");
        return -1;
    }

    if (rl.rlim_cur < limit) {
        rl.rlim_cur = limit;
        if (setrlimit(RLIMIT_NOFILE, &rl) == -1) {
            perror("setrlimit failed");
            return -1;
        }
    }

    return 0;
}

int main(int argc, char** argv)
{
    std::string configFile = QWEN2_CONFIG_FILE;
    if (argc == 2) {
        configFile = std::string(argv[1]);
    }

    if (setFdLimit(4096)) {
        return -1;
    }

    // Register the signal handler
    std::signal(SIGINT, signalHandler);
    std::signal(SIGTSTP, signalHandler);

    const std::string splitStr = "----------------------------------------------------------------------------------";

    try {
        std::string inputPrompt;
        std::string rolePrompt = "你是一个智能助理.";
        int roleFlag = 0;

        gQwen2Client = std::make_shared<Qwen2Client>();
        gQwen2Client->setconfig(configFile);

        std::map<int, std::string> options = gQwen2Client->getOptionalPrompts();
        std::map<int, std::string> roleOptions = gQwen2Client->getOptionalRolePrompts();

        while (!Qwen2Client::mExited.load()) {
            inputPrompt.clear();

            if (options.size()) {
                if (roleFlag == 0) {
                    std::cout << std::endl << splitStr << std::endl;
                    for (const auto& prompt : options) {
                        if ((prompt.second == "Role setting") && roleOptions.size()) {
                            std::cout << prompt.first << ": " << prompt.second << ": "<< rolePrompt << std::endl;
                            std::cout << splitStr << std::endl;
                        } else {
                            std::cout << prompt.first << ": " << prompt.second << std::endl;
                        }
                    }
                    std::cout << options.size() << USER_INPUT << std::endl;
                    std::cout << splitStr << std::endl;
                }
                std::cout << "[YOU]: ";

                int choice;
                std::string strChoice;
                std::getline(std::cin, strChoice);

                // 检查是否仅仅是回车
                if (strChoice.empty()) {
                    continue;
                }

                if (strChoice == "exit") {
                    std::cout << "Qwen2 退出" << std::endl;
                    gQwen2Client->exitClient();
                    break;
                }

                char extraChar;
                std::istringstream iss(strChoice);
                if (iss >> choice && !(iss >> extraChar)) {
                    std::map<int, std::string>::iterator it;
                    if (roleFlag == 0) {
                        it = options.find(choice);
                    } else {
                        // role setting get
                        it = roleOptions.find(choice);
                        if (it != roleOptions.end()) {
                            rolePrompt = it->second;
                            roleFlag = 0;
                            continue;
                        } else {
                            std::cout << "Role setting not correct, please continue setting role: " << std::endl;
                            continue;
                        }
                    }

                    if (it != options.end()) {
                        inputPrompt = it->second;
                        std::cout << "[YOU]: " << inputPrompt << std::endl;

                        // role setting input
                        if (inputPrompt == "Role setting") {
                            if (!roleOptions.size()) {
                                std::cout << "Role options dosn't exit, please check config.json" << std::endl;
                                roleFlag = 0;
                                continue;
                            }

                            for (const auto& prompt : roleOptions) {
                                std::cout << prompt.first << ": " << prompt.second << std::endl;
                            }
                            roleFlag = 1;
                            continue;
                        }
                    } else {
                        if (choice == options.size()) {
                            std::cout << "[YOU]: ";
                            // inputPrompt = getChineseInput();
                            std::getline(std::cin, inputPrompt);
                            if (inputPrompt.empty()) {
                                continue;
                            }
                            if (inputPrompt == "exit") {
                                std::cout << "Qwen2 退出" << std::endl;
                                gQwen2Client->exitClient();
                                break;
                            }
                        } else {
                            continue;
                        }
                    }
                } else {
                    iss.clear();
                    continue;
                }
            } else {
                std::cout << std::endl << "[YOU]: ";
                std::getline(std::cin, inputPrompt);

                if (inputPrompt.empty()) {
                    continue;
                }

                if (inputPrompt == "exit") {
                    std::cout << "Qwen2 退出" << std::endl;
                    gQwen2Client->exitClient();
                    break;
                }
            }

            // Generate response text
            std::cout << "[Qwen2]: ";
            std::string generatedText = gQwen2Client->genResponse(rolePrompt, inputPrompt);
        }
    } catch (const std::exception& e) {
        return 1;
    }

    return 0;
}
